;;; -*- TI-Asm -*-
;;;
;;; Mimas - Assembly language IDE for the TI-83 Plus
;;; 
;;; Copyright (C) 2010-2012 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; GetProgramSymbolOffset:
;;
;; Get the offset to a given symbol in the current program file, and
;; check if that offset is valid.
;;
;; Input:
;; - HL = symbol number
;;
;; Output:
;; - DE = symbol number
;; - HL = offset to symbol in program file (if valid)
;; - Carry flag clear if symbol number is out of bounds
;;
;; Destroys:
;; - F

P1_GetProgramSymbolOffset:
	ex de,hl
P1_GetProgramSymbolOffsetDE:
	ld hl,(curProgSymbolCount)
	scf
	sbc hl,de
	ret c
	ld hl,(curProgHeader + PROGHEADER_SYMBOL_START)
	add hl,de
	add hl,de
	add hl,de
	ret

;; GetProgramSymbolPtrWrite:
;;
;; Get the address of the given symbol's info block.  The current
;; program will be moved into RAM if necessary.
;;
;; Input:
;; - HL = symbol number
;;
;; Output:
;; - HL = address of symbol info (3 bytes: string offset L, string
;;   offset H + flags, ref count)
;;
;; Destroys:
;; - AF

P1_GetProgramSymbolPtrWrite:
	push de
	 call P1_GetProgramSymbolOffset
	 call c,P1_Error_Bytecode
	 ld a,(curProgStartPage)
	 or a
	 jr z,P1_ProgramOffsetToAddress_1
	 push bc
	  ld bc,0
	  call P1_ProgramInsertDeleteMem
	  ex de,hl
	  pop bc
	 pop de
	ret

;; ProgramOffsetToAddress:
;;
;; Get the absolute address of a given offset in the input file.
;;
;; Input:
;; - HL = offset into program file
;;
;; Output:
;; - AHL = absolute address
;;
;; Destroys:
;; - F

P1_ProgramOffsetToAddress:
	push de
P1_ProgramOffsetToAddress_1:
	 ex de,hl
	 ld hl,(curProgStartPtr)
	 ld a,(curProgStartPage)
	 call P1_Add_AHL_DE
	 pop de
	ret

;; InsDelProgramCodeMem:
;;
;; Insert or delete memory into the current program section.  The
;; current program will be unarchived if necessary.
;;
;; Input:
;; - HL = offset into program file to location where memory should be
;;   inserted, or to location of the first byte to be deleted (must be
;;   between (curSectionHeader + SECTHEADER_DATA_START) and
;;   (curSectionHeader + SECTHEADER_DATA_END), inclusive)
;; - BC = number of bytes to insert (negative number to delete)
;;
;; Output:
;; - DE = address of first byte inserted
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - AF, HL

P1_InsDelProgramCodeMem:
	call P1_ProgramInsertDeleteMem
	push de

	 ;; Update global pointers
	 call P1_ProgramInsDelUpdateHeaderForCode

	 ;; Update this section table entry
	 ld hl,(curSectionHeader + SECTHEADER_DATA_END)
	 add hl,bc
	 ld (curSectionHeader + SECTHEADER_DATA_END),hl

	 push bc
	  ld hl,(curSectionOffset)
	  call P1_ProgramOffsetToAddress
	  ex de,hl
	  ld hl,curSectionHeader
	  ld bc,SECTHEADER_LENGTH
	  ldir
	  pop bc

	 ;; Update later section table entries
	 ld hl,(curSectionOffset)
	 call P1_ProgramInsDelUpdateSTab
	 pop de
	ret

;; InsDelProgramStringMem:
;;
;; Insert or delete memory into the current program's string area.
;; The current program will be unarchived if necessary.
;;
;; Input:
;; - HL = offset into program file to location where memory should be
;;   inserted, or to location of the first byte to be deleted (must be
;;   between (curProgHeader + PROGHEADER_SYMBOL_STRING_START) and
;;   (curProgHeader + PROGHEADER_SYMBOL_STRING_END), inclusive)
;; - BC = number of bytes to insert (negative number to delete)
;;
;; Output:
;; - DE = address of first byte inserted
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - AF, HL

P1_InsDelProgramStringMem:
	push hl
	 call P1_ProgramInsertDeleteMem
	 pop hl
	push de
	 push hl
	  ;; Update global pointers
	  call P1_ProgramInsDelUpdateHeaderForStrings
	  pop hl
	 push bc

	  ;; Update symbols (symbol string pointers are relative to the
	  ;; start of string memory)
	  ld de,(curProgHeader + PROGHEADER_SYMBOL_STRING_START)
	  or a
	  sbc hl,de		; HL = offset into string memory
	  ld b,h
	  ld c,l

	  ;; get address of start of symbol table (this will throw an
	  ;; error if the symbol table is empty - a program's symbol
	  ;; table should never, in fact, be empty at the time this
	  ;; function is called)
	  ld hl,0
	  call P1_GetProgramSymbolPtrWrite
	  ex de,hl

	  ld hl,(curProgSymbolCount)

	  ;; DE = address of start of symbol table
	  ;; HL = number of symbols left to check
	  ;; BC = position where bytes were allocated
	  ;; (SP) = number of bytes allocated

P1_InsDelProgramStringMem_SymbolLoop:
	  inc de		; -> offset H
	  inc de		; -> ref count
	  ld a,(de)		; if symbol is free, do nothing
	  or a
	  jr z,P1_InsDelProgramStringMem_SymbolSkip

	  ;; if symbol's string ptr >= BC, then we need to adjust it
	  dec de		; -> offset H
	  dec de		; -> offset L
	  ld a,(de)
	  sub c
	  inc de		; -> offset H
	  ld a,(de)
	  sbc a,b
	  jr c,P1_InsDelProgramStringMem_SymbolNoAdjust

	  ex (sp),hl
	  dec de		; -> offset L
	  ld a,(de)
	  add a,l
	  ld (de),a
	  inc de		; -> offset H
	  ld a,(de)
	  adc a,h
	  ld (de),a
	  ex (sp),hl

P1_InsDelProgramStringMem_SymbolNoAdjust:
	  inc de		; -> ref count
P1_InsDelProgramStringMem_SymbolSkip:
	  inc de		; -> next symbol offset L
	  dec hl
	  ld a,h
	  or l
	  jr nz,P1_InsDelProgramStringMem_SymbolLoop
	  pop bc
	 pop de
	ret

;; AppendProgramSymbol:
;;
;; Allocate a fresh symbol table entry in the current program.  The
;; current program will be unarchived if necessary.
;;
;; Output:
;; - DE = address of first byte inserted
;; - BC = 3
;;
;; Destroys:
;; - AF, HL

P1_AppendProgramSymbol:
	ld hl,(curProgHeader + PROGHEADER_SYMBOL_END)
	ld bc,3
	call P1_ProgramInsertDeleteMem
	;; Update global pointers
	jr P1_ProgramInsDelUpdateHeaderForSyms

;; InsDelProgramImportMem:
;;
;; Insert or delete memory into the current program's imports area.
;; The current program will be unarchived if necessary.
;;
;; Input:
;; - HL = offset into program file to location where memory should be
;;   inserted, or to location of the first byte to be deleted (must be
;;   between (curProgHeader + PROGHEADER_IMPORT_START) and
;;   (curProgHeader + PROGHEADER_IMPORT_END), inclusive)
;; - BC = number of bytes to insert (negative number to delete) (must
;;   be a multiple of 9)
;;
;; Output:
;; - DE = address of first byte inserted
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - AF, HL

P1_InsDelProgramImportMem:
	call P1_ProgramInsertDeleteMem
	;; Update global pointers
	jr P1_ProgramInsDelUpdateHeaderForImports

;; InsDelProgramSectionMem:
;;
;; Insert or delete memory into the current program's section table
;; area.  The current program will be unarchived if necessary.
;;
;; Note: (curSectionOffset) will not be updated to reflect the new
;; offset of the current section, and thus should be considered
;; destroyed.
;;
;; Input:
;; - HL = offset into program file to location where memory should be
;;   inserted, or to location of the first byte to be deleted (must be
;;   between (curProgHeader + PROGHEADER_SECTION_TABLE_START) and
;;   (curProgHeader + PROGHEADER_SECTION_TABLE_END), inclusive)
;; - BC = number of bytes to insert (negative number to delete) (must
;;   be a multiple of SECTHEADER_LENGTH)
;;
;; Output:
;; - DE = address of first byte inserted
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - AF, HL
;; - (curSectionOffset)

P1_InsDelProgramSectionMem:
	call P1_ProgramInsertDeleteMem
	;; Update global pointers
	jr P1_ProgramInsDelUpdateHeaderForLateSections

;; ProgramInsDelUpdateHeaderForCode:
;;
;; Update program header pointers after inserting program code memory.
;; Program is assumed to be in RAM.
;;
;; Input:
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - F, HL

P1_ProgramInsDelUpdateHeaderForCode:
	ld hl,(curProgHeader + PROGHEADER_SYMBOL_START)
	add hl,bc
	ld (curProgHeader + PROGHEADER_SYMBOL_START),hl
P1_ProgramInsDelUpdateHeaderForSyms:
	ld hl,(curProgHeader + PROGHEADER_SYMBOL_STRING_START)
 	add hl,bc
	ld (curProgHeader + PROGHEADER_SYMBOL_STRING_START),hl
P1_ProgramInsDelUpdateHeaderForStrings:
	ld hl,(curProgHeader + PROGHEADER_CONSTANT_START)
 	add hl,bc
	ld (curProgHeader + PROGHEADER_CONSTANT_START),hl
	ld hl,(curProgHeader + PROGHEADER_IMPORT_START)
 	add hl,bc
	ld (curProgHeader + PROGHEADER_IMPORT_START),hl
P1_ProgramInsDelUpdateHeaderForImports:
	ld hl,(curProgHeader + PROGHEADER_SECTION_TABLE_START)
	add hl,bc
	ld (curProgHeader + PROGHEADER_SECTION_TABLE_START),hl
P1_ProgramInsDelUpdateHeaderForEarlySections:
	ld hl,(curSectionOffset)
	add hl,bc
	ld (curSectionOffset),hl
P1_ProgramInsDelUpdateHeaderForLateSections:
	ld hl,(curProgHeader + PROGHEADER_SECTION_TABLE_END)
	add hl,bc
	ld (curProgHeader + PROGHEADER_SECTION_TABLE_END),hl
P1_ProgramModifyHeaderNoArc:
	push bc
	 push de
	  ld hl,curProgHeader
	  ld de,(curProgStartPtr)
	  ld bc,PROGHEADER_LENGTH
	  ldir
	  pop de
	 pop bc
	ret

;; ProgramInsDelUpdateSTab:
;;
;; Update section table offsets after inserting program code memory.
;;
;; Input:
;; - HL = offset to section into which bytes were inserted
;; - BC = number of bytes inserted
;;
;; Destroys:
;; - AF, DE, HL

P1_ProgramInsDelUpdateSTab:
	ld de,SECTHEADER_LENGTH + SECTHEADER_DATA_START
P1_ProgramInsDelUpdateSTab_Loop:
	add hl,de
	ld de,(curProgHeader + PROGHEADER_SECTION_TABLE_END)
	or a
	sbc hl,de
	ret nc
	add hl,de
	push hl
	 call P1_ProgramOffsetToAddress
	 ;; update section start
	 ld a,(hl)
	 add a,c
	 ld (hl),a
	 inc hl
	 ld a,(hl)
	 adc a,b
	 ld (hl),a
	 inc hl
	 ;; update section end
	 ld a,(hl)
	 add a,c
	 ld (hl),a
	 inc hl
	 ld a,(hl)
	 adc a,b
	 ld (hl),a
	 pop hl
	ld de,SECTHEADER_LENGTH
	jr P1_ProgramInsDelUpdateSTab_Loop

;; ProgramInsertDeleteMem:
;;
;; Insert or delete memory from the current program file.  The current
;; program will be unarchived if necessary.  This routine will throw
;; an error if there is not enough RAM available to perform the
;; operation.
;;
;; (If BC = 0, no memory is inserted or deleted, but the program is
;; still unarchived if necessary.  BC = 0 is the only time when it is
;; appropriate to call this routine directly, rather than using one of
;; the InsDelProgram*Mem routines.)
;;
;; Input:
;; - HL = offset to first byte to insert/delete
;; - BC = amount to insert / negative amount to delete
;;
;; Output:
;; - DE = address of first inserted byte / address of first byte
;;   following deleted mem
;; - If memory was inserted, it is uninitialized
;;
;; Destroys:
;; - AF, HL

P1_ProgramInsertDeleteMem_Archived:
	push hl
	 push bc
	  push ix
	   rst rPUSHREALO1

	   ld hl,P1_ProgramInsertDeleteMem_Error
	   call APP_PUSH_ERRORH

	    ld hl,curProgFileName
	    rst rMOV9TOOP1
	    BCALL _ChkFindSym

	    ;; if file doesn't exist, or is not currently archived, this
	    ;; is really bad and we should probably abort immediately
	    ccf
	    sbc a,a
	    and b
	    jr nz,P1_ProgramInsertDeleteMem_ArchiveOK
	    BCALL _ErrDomain
P1_ProgramInsertDeleteMem_ArchiveOK:

	    ;; mark variable as having been unarchived
	    set 7,(hl)

	    ;; use UnarchiveVar rather than Arc_Unarc since the latter
	    ;; destroys all of OP1 and OP3
	    BCALL _UnarchiveVar

	    call APP_POP_ERRORH

	   BCALL B_ReloadProgram
	   BCALL _PopRealO1
	   pop ix
	  pop bc
	 pop hl
P1_ProgramInsertDeleteMem:
	ld a,(curProgStartPage)
	or a
	jr nz,P1_ProgramInsertDeleteMem_Archived
	ex de,hl
	;; check that insertion offset is sane
	ld hl,(curProgStartPtr)
	dec hl
	dec hl
	ld a,(hl)
	sub e
	inc hl
	ld a,(hl)
	inc hl
	sbc a,d
	jr nc,P1_ProgramInsertDeleteMem_ValidOffset
	;; if we ever get here, it's a bug
	BCALL _ErrDimension
	;; UNREACHABLE

P1_ProgramInsertDeleteMem_ValidOffset:
	add hl,de
	push bc
	 ld a,b
	 add a,a
	 jr c,P1_ProgramInsertDeleteMem_Delete
	 or c
	 jr z,P1_ProgramInsertDeleteMem_DoneHL
	 call P1_DoInsertMem
	 jr P1_ProgramInsertDeleteMem_Done
P1_ProgramInsertDeleteMem_Delete:
	 xor a
	 sub c
	 ld e,a
	 sbc a,a
	 sub b
	 ld d,a
	 push hl
	  BCALL _DelMem
	  pop hl
P1_ProgramInsertDeleteMem_DoneHL:
	 ex de,hl
P1_ProgramInsertDeleteMem_Done:
	 pop bc
	;; update program size
	ld hl,(curProgStartPtr)
	dec hl
	dec hl
	ld a,(hl)
	add a,c
	ld (hl),a
	inc hl
	ld a,(hl)
	adc a,b
	ld (hl),a
	ret

P1_Error_Bytecode:
	ld hl,emsg_BytecodeError
	jr P1_ThrowGeneralError
P1_ProgramInsertDeleteMem_Error:
	cp E_Memory
	jr nz,P1_JError
	ld hl,emsg_CannotUnarchive
P1_ThrowGeneralError:
	ld a,E_AppErr1 & 7Fh
	ld (customErrorStr),hl
;	ld (customErrorParam),de
P1_JError:
	BCALL _JError
	;; UNREACHABLE
