;;; -*- TI-Asm -*-
;;;
;;; Mimas plugin to display a scroll bar in the assembly editor
;;;
;;; Copyright (C) 2012 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

	processor z80
	nolist
	include <ti83plus.inc>
	list

MIMAS_PLUGIN equ 1

	nolist
	include <mimas.inc>
	include <defs.inc>
	include <insts.inc>
	include <plugapi.inc>
	list

;; PLUGIN_SCROLL:
;;
;; This plugin is called in two cases:
;;
;; - After DisplayAsmEditor, the plugin is called with L = 0 and DE =
;;   offset to end of window.  The plugin uses this information to
;;   update the scroll bar state.
;;
;; - Before GetKeyCursor, the plugin is called with L = 1.  The plugin
;;   sets up its own cursor hook and waits for a keypress; it then
;;   returns with the carry flag clear.
;;
;; Input:
;; - L = 0 and DE = end of window to update scroll bar state
;; - L = 1 to display scroll bar and wait for a keypress
;;
;; Output:
;; - (L = 1 case) A = keycode and carry flag clear
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1-OP3

    BEGIN_PLUGIN PLUGIN_SCROLL

	dec l
	jr z,GetKeyScrollBar

;; UpdateScrollBar:
;;
;; Update scroll bar state after the window has been scrolled.
;;
;; Input:
;; - (curSectionHeader + SECTHEADER_DATA_START) = start of section
;; - (editStartOffset) = start of window
;; - DE = end of window
;; - (curSectionHeader + SECTHEADER_DATA_END) = end of section
;;
;; Output:
;; - (editScrollBarThumbSize) = size of thumb
;; - (editScrollBarThumbPos) = position of thumb (offset from bottom
;;   of bar)
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1-OP3

UpdateScrollBar:
	;; check if the blank line at the end of the section is visible
	ld hl,(plotSScreen + 55*12)
	ld a,h
	or l
	ld hl,(curSectionHeader + SECTHEADER_DATA_END)
	jr z,UpdateScrollBar_EndOfSection
	;; add 3 bytes to represent the blank line
	inc hl
	inc hl
	inc hl
UpdateScrollBar_EndOfSection:
	sbc hl,de
	push hl			; size of data after window
	 ld hl,(editStartOffset)
	 ld de,(curSectionHeader + SECTHEADER_DATA_START)
	 sbc hl,de
	 push hl		; size of data before window
	  ld hl,(curSectionHeader + SECTHEADER_DATA_END)
	  sbc hl,de		; total size of section
	  jr z,UpdateScrollBar_Disable2

	  ;; Set thumb size based on total section size, with an
	  ;; estimate of 24 bytes per page

	  ;; (size of the thumb should give a clue as to how large the
	  ;; section is, but the thumb size should not change as we
	  ;; scroll through the file)
	  push hl
	   ld hl,57 * 24
	   BCALL _SetXXXXOP2
	   BCALL _OP2ToOP1
	   pop hl
	  BCALL _SetXXXXOP2
	  BCALL _FPDiv
	  BCALL _ConvOP1
	  inc a
	  ld (editScrollBarThumbSize),a

	  ;; Set thumb position by dividing the remaining pixels
	  ;; according to the ratio between the chunk of data above
	  ;; the window, and the chunk of data below the window

	  ;; (when we are at the top/bottom of the section, the thumb
	  ;; should be placed precisely at the top/bottom of the bar)
	  ld a,57
	  sub e
	  BCALL _SetXXOP1
	  pop hl
	 push hl
	  BCALL _SetXXXXOP2
	  BCALL _FPMult
UpdateScrollBar_Disable2:
	  pop hl
	 pop de
	;; disable scroll bar if entire section is visible on screen
	;; (which includes the case where the section is empty)
	or a
	adc hl,de
	jr z,DisableScrollBar
	BCALL _SetXXXXOP2
	BCALL _FPDiv
	BCALL _ConvOP1
	ld a,57
	sub e
	ld (editScrollBarThumbPos),a
 	ret

DisableScrollBar:
	res editScrollBar,(iy + asm_Flag2)
	ret

;; GetKeyScrollBar:
;;
;; Get a keypress, displaying a blinking small-font cursor.  If
;; desired, also show a scroll bar at the right edge of the screen.
;; The scroll bar is shown only for ~400 milliseconds, and will be
;; hidden when any key is pressed other than the up or down arrow
;; keys.  While the scroll bar is displayed, the shift indicator (if
;; any) is hidden.
;;
;; Input:
;; - (penCol) = X coordinate of cursor plus 1
;; - (penRow) = Y coordinate of cursor
;; - editScrollBar, (iy + asm_Flag2) = 1 to display scroll bar
;; - (editScrollBarThumbPos) = position of thumb
;; - (editScrollBarThumbSize) = size of thumb
;; - (curScreenBuffer) = screen buffer
;;
;; Output:
;; - A = keycode
;; - Carry flag clear
;;
;; Destroys:
;; - F, BC, DE, HL
;; - curRow, curCol

GetKeyScrollBar:
	in a,(6)
	dec a
	out (6),a
	ld hl,ScrollBarCursorHook
	call P1_GetKeyCursor_CustomHook
	or a
	jp RAMCodeStubReturn

ScrollBarCursorHook:
	db 83h
	bit editScrollBar,(iy + asm_Flag2)
	jr z,ScrollBarCursorHook_NoScrollBar

	;; when cursor is initially displayed, show scroll bar
	ld c,1111b
	cp 22h
	jr z,ScrollBarCursorHook_PutScrollBar

	;; hide scroll bar when cursor is hidden or shift state is
	;; changed, *except* when GetKey has already finished (curAble
	;; is cleared) *and* an up/down arrow key was pressed, in
	;; which case leave the scroll bar in place
	bit curAble,(iy + curFlags)
	jr nz,ScrollBarCursorHook_EraseScrollBar
	ld b,a
	ld a,(kbdGetKy)
	and ~5
	ld a,b
	ret z
ScrollBarCursorHook_EraseScrollBar:
	call DisableScrollBar
	ld c,0
ScrollBarCursorHook_PutScrollBar:
	push af
	 push ix
	  di
	  ld a,80h
	  call 000Bh
	  out (10h),a
	  ld hl,(curScreenBuffer)
	  dec hl
	  ld de,12
	  ld ix,editScrollBarThumbPos
	  ld a,2Bh
	  call 000Bh
	  out (10h),a
	  ld b,57
	  ld a,c
	  jr ScrollBarCursorHook_PutScrollBarStart
ScrollBarCursorHook_PutScrollBarLoop:
	  ld a,(ix)		; editScrollBarThumbPos
	  sub b
	  cp (ix+1)		; editScrollBarThumbSize
	  sbc a,a
	  or 1001b
ScrollBarCursorHook_PutScrollBarStart:
	  add hl,de
	  xor (hl)
	  and c
	  xor (hl)
	  call 000Bh
	  out (11h),a
	  djnz ScrollBarCursorHook_PutScrollBarLoop
	  ei
	  pop ix
	 pop af
	;; Z -> scroll bar displayed
	;; (skip displaying shift state but display cursor as usual)
	jp z,P1_SmallCursorHook_CursorOnly

	;; NZ -> scroll bar hidden
	;; (display shift state and show/hide cursor as usual)
ScrollBarCursorHook_NoScrollBar:
	jp P1_SmallCursorHook + 1

    END_PLUGIN
